//----------------------------------------------------------------------------
//  Project TCPDUMP Filter
//
//  Copyright  1999. All Rights Reserved.
//
// SUBSYSTEM:           TCPDUMP Filter Application
// FILE:                tcpdfilter.cpp
// AUTHOR:              Trenton D. Adams
// PLATFORM:            Linux
// COMPILER:            GCC for Linux
//
//----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HEXOUT 0x01
#define PACKETDECODE 0x02
#define HELP 0x80000000

typedef struct 
{
  char word1[5],
    word2[5],
    word3[5],
    word4[5],
    word5[5],
    word6[5],
    word7[5],
    word8[5];
} THexString;

typedef struct
{
  char version,   // IP version
    headerLength, // header length
    flags,        // flags
    ttl,          // time to live
    protocol;     // protocol

  long sourceIP,   // source ip address
    destIP;    // destination ip address

  short tos,      // type of service
    totalLength,  // total packet length
    id,           // packet id
    fragOffset,   // fragmentation offset
    headChecksum; // header checksum
} TIPHeader;

int processCmdLine (int argc, char *argv[]);
void usage(void);
int stringConvert(char string[255], char retString[255]);
int getHeader(THexString hexString[2], TIPHeader *ipHeader);
void displayHeader(TIPHeader ipHeader);

int main(int argc, char *argv[])
{
  char tempString[255],
    retString[3];

  THexString hexString,
    hexStringArray[2];
  TIPHeader ipHeader;

  int commandOptions,
    x, 
    lineCount = 0;

  commandOptions = processCmdLine (argc, argv);

  if (commandOptions & 0x80000000)
    {
      usage();
      exit (1);
    }

  while (fgets(tempString, sizeof(tempString), stdin) != NULL)
    {  // begin while
       if ((tempString[0] == '\t') && (tempString[1] == '\t') && (tempString[2] == '\t'))
	 {
	   lineCount++;
	   // tcpdump -x uses three tabs a space, and then the hex values
	   sscanf(tempString, "\t\t\t %s %s %s %s %s %s %s %s",
		  &hexString.word1, &hexString.word2,
		  &hexString.word3, &hexString.word4,
		  &hexString.word5, &hexString.word6,
		  &hexString.word7, &hexString.word8);
	   
	   if (lineCount == 1)
	     hexStringArray[0] = hexString;
	   else if (lineCount == 2)
	     {
	       hexStringArray[1] = hexString;
	       getHeader (hexStringArray, &ipHeader);
	     }

	   // convert each hex string into two ASCII characters
	   stringConvert(hexString.word1, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word2, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word3, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word4, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word5, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word6, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word7, retString);
	   printf ("%s", retString);
	   stringConvert(hexString.word8, retString);
	   printf ("%s", retString);
	   
	   if (commandOptions & HEXOUT)
	     {
	       // get rid of the tabs and put spaces in
	       tempString[0] = 32;
	       tempString[1] = 32;
	       tempString[2] = ':';
	       printf ("%s", tempString);
	     }
	   else
	     printf ("\n");
	   
	   for(x = 0;x < 5; x++)
	     {
	       tempString[x]  = 48;
	       hexString.word1[x] = 48;
	       hexString.word2[x] = 48;
	       hexString.word3[x] = 48;
	       hexString.word4[x] = 48;
	       hexString.word5[x] = 48;
	       hexString.word6[x] = 48;
	       hexString.word7[x] = 48;
	       hexString.word8[x] = 48;

	     }
	 }
	else
	  {
	    if (commandOptions & PACKETDECODE)
	      displayHeader (ipHeader);
	    // Reaching here probably means a new packet has started
	    lineCount = 0;
	      printf ("\n\n%s\n", tempString);
	    fflush (stdout);
	  }
     } // End While

   return (0);
}

int processCmdLine (int argc, char *argv[])
/* Function:   Processes the command line arguments
   Input:      argc, argv
   Output:     none
   Return:     bit 0  set = hex output
               bit 1  set = decode packet
	       bit 31 set = display usage message
*/
{ // begin processCmdLine
  int index,
    valueToReturn = 0;
  
  if (argc > 1)
    for (index = 1; index < argc; index++)
      {
	if (!strcmp (argv[index], "-x"))
	  valueToReturn |= HEXOUT; // hex output
	else if (!strcmp(argv[index], "-d"))
	  valueToReturn |= PACKETDECODE; // decode packet
	else
	  valueToReturn |= HELP; // display usage message
      }

  return (valueToReturn);
} // end processCmdLine


void usage(void)
{ // begin usage ()
  printf ("\n\nTCPDUMP Filter Usage\t\t (tcpdfilter)\n\n"
	  "\ttcpdump -x -l -n | tcpdfilter [-x] [-d]\n"
	  "\t-x  = include hex output\n"
	  "\t-d  = decode packets\n\n\n");
} // end usage ()

int stringConvert(char string[5], char retString[3])
/* Function:   Converts two bytes of HEX (ADFB) to two bytes of ASCII
   Input:      string with the HEX bytes and a NULL character
               retString buffer for the return string with two ASCII
               characters and a NULL
   Output:     none
   Return:     errorCode = 0 = no errors
               errorCode = 1 = errors with input string

   ALGORITHM
      Take apart the two bytes (actually four) of hex values and stick
      them in two separate null terminated strings
         sscanf each string for hex values
         put returned ASCII character in proper position of retString
         repeat step for next string
*/
{ // Begin stringConvert ()
   int character;

   char hex1[3];
   char hex2[3];

   hex1[0] = string[0];
   hex1[1] = string[1];
   hex1[2] = 0;

   hex2[0] = string[2];
   hex2[1] = string[3];
   hex2[2] = 0;

   sscanf (hex1, "%x", &character);
   if ((character >= 32) && (character < 127))
      retString[0] = (char)character;
   else
     retString[0] = 46; // a '.'

   sscanf ( hex2, "%x", &character);
   if ((character >= 32) && (character < 127))
      retString[1] = (char)character;
   else
     retString[1] = 46; // a '.'

   retString[2] = 0;

   return (0);
} // end stringConvert ()


int getHeader(THexString hexString[2], TIPHeader *ipHeader)
/* Function:   Converts hex values stored in the array of hexStrings
               to an IP header;
   Input:      strings with the HEX bytes
               ipHeader buffer for the return ipHeader structure
   Output:     none
   Return:     errorCode = 0 = no errors
               errorCode = 1 = errors with input string

*/
{ // Begin getHeader ()
  unsigned long tempInt;

  char hex32[9];

  sscanf (hexString[0].word1, "%x", &tempInt);
  ipHeader->version = (tempInt & 0xF000) >> 12;
  ipHeader->headerLength = (tempInt & 0x0F00) >> 8;
  ipHeader->tos = (tempInt & 0X00FF);

  sscanf (hexString[0].word2, "%x", &tempInt);
  ipHeader->totalLength = tempInt;

  sscanf (hexString[0].word3, "%x", &tempInt);
  ipHeader->id = tempInt;

  sscanf (hexString[0].word4, "%x", &tempInt);
  ipHeader->flags = (tempInt & 0xe000) >> 13;
  ipHeader->fragOffset = tempInt & 0x1FFF;

  sscanf (hexString[0].word5, "%x", &tempInt);
  ipHeader->ttl = (tempInt & 0x0FF00) >> 8;
  ipHeader->protocol = tempInt & 0x00FF;

  sscanf (hexString[0].word6, "%x", &tempInt);
  ipHeader->headChecksum = tempInt;

  strcpy (hex32, hexString[0].word7);
  strcat (hex32, hexString[0].word8);
  sscanf (hex32, "%x", &tempInt);
  ipHeader->sourceIP = tempInt;

  strcpy (hex32, hexString[1].word1);
  strcat (hex32, hexString[1].word2);
  sscanf (hex32, "%x", &tempInt);
  ipHeader->destIP = tempInt;

} // end getHeader ()

void displayHeader(TIPHeader ipHeader)
/* Function:   Displays the packet header information
   Input:      ipHeader
   Output:     none
   Return:     none
*/
{ // begin displayHeader ()
  printf ("\n\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n"
	  "IP packet header\n"
	  "IP Version : %d\n"
	  "IP Header Length (32 bit quantities) : %d\n"
	  "Service Type\n"
	  "\tPrecedence : %d\n"
	  "\tType of Service : %d\n"
	  "IP Packet Length (bytes) : %d\n"
	  "Packet Identifier : %ld\n"
	  "Flags\n"
	  "\tAllow Fragmentation? : %s\n"
	  "\tLast Fragment? : %s\n"
	  "Fragment Offset : %d\n"
	  "Time to Live (TTL) : %u\n"
	  "Protocol Number : %d\n"
	  "Source IP Address : %d.%d.%d.%d\n"
	  "Destination IP Address : %d.%d.%d.%d\n"
	  "----------------------------------------\n",
	  ipHeader.version, 
	  ipHeader.headerLength,
	  (ipHeader.tos & 0xE0) >> 5,
	  ipHeader.tos & 0x1F,
	  (unsigned short)ipHeader.totalLength,
	  (unsigned short)ipHeader.id,
	  (ipHeader.flags & 0x02) ? "No":"Yes",
	  (ipHeader.flags & 0x01) ? "No":"Yes",
	  ipHeader.fragOffset,
	  (unsigned char)ipHeader.ttl,
	  ipHeader.protocol,
	  (ipHeader.sourceIP & 0xFF000000) >> 24,
	  (ipHeader.sourceIP & 0x00FF0000) >> 16,
	  (ipHeader.sourceIP & 0x0000FF00) >> 8,
	  (ipHeader.sourceIP & 0x000000FF),
	  (ipHeader.destIP & 0xFF000000) >> 24,
	  (ipHeader.destIP & 0x00FF0000) >> 16,
	  (ipHeader.destIP & 0x0000FF00) >> 8,
	  (ipHeader.destIP & 0x000000FF));
} // end displayHeader ()
